# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import warnings

from ctypes import (
    c_byte,
    c_char_p,
    c_float,
    c_int,
    c_size_t,
    c_uint,
    c_uint8,
    c_void_p,
    py_object,
    CFUNCTYPE,
    POINTER,
    Structure,
)

warnings.warn(
    "Importing from numba.cuda.cudadrv.drvapi is deprecated.",
    DeprecationWarning,
    stacklevel=2,
)

cu_device = c_int
cu_device_attribute = c_int  # enum
cu_context = c_void_p  # an opaque handle
cu_module = c_void_p  # an opaque handle
cu_jit_option = c_int  # enum
cu_jit_input_type = c_int  # enum
cu_function = c_void_p  # an opaque handle
cu_device_ptr = c_size_t  # defined as unsigned long long
cu_stream = c_void_p  # an opaque handle
cu_event = c_void_p
cu_link_state = c_void_p
cu_function_attribute = c_int
cu_uuid = c_byte * 16  # Device UUID

cu_stream_callback_pyobj = CFUNCTYPE(None, cu_stream, c_int, py_object)

cu_occupancy_b2d_size = CFUNCTYPE(c_size_t, c_int)


# Mirrors the definition of CUipcMemHandle in cuda.h
class cu_ipc_mem_handle(Structure):
    _fields_ = [("reserved", c_uint8 * 64)]


# See https://docs.nvidia.com/cuda/cuda-driver-api/group__CUDA__TYPES.html
CU_STREAM_DEFAULT = 0
CU_STREAM_LEGACY = 1
CU_STREAM_PER_THREAD = 2

API_PROTOTYPES = {
    # CUresult cuInit(unsigned int Flags);
    "cuInit": (c_int, c_uint),
    # CUresult cuDriverGetVersion (int* driverVersion )
    "cuDriverGetVersion": (c_int, POINTER(c_int)),
    # CUresult cuDeviceGetCount(int *count);
    "cuDeviceGetCount": (c_int, POINTER(c_int)),
    # CUresult cuDeviceGet(CUdevice *device, int ordinal);
    "cuDeviceGet": (c_int, POINTER(cu_device), c_int),
    # CUresult cuDeviceGetName ( char* name, int  len, CUdevice dev )
    "cuDeviceGetName": (c_int, c_char_p, c_int, cu_device),
    # CUresult cuDeviceGetAttribute(int *pi, CUdevice_attribute attrib,
    #                               CUdevice dev);
    "cuDeviceGetAttribute": (
        c_int,
        POINTER(c_int),
        cu_device_attribute,
        cu_device,
    ),
    # CUresult cuDeviceComputeCapability(int *major, int *minor,
    #                                    CUdevice dev);
    "cuDeviceComputeCapability": (
        c_int,
        POINTER(c_int),
        POINTER(c_int),
        cu_device,
    ),
    # CUresult cuDevicePrimaryCtxGetState(
    #              CUdevice dev,
    #              unsigned int* flags,
    #              int* active)
    "cuDevicePrimaryCtxGetState": (
        c_int,
        cu_device,
        POINTER(c_uint),
        POINTER(c_int),
    ),
    # CUresult cuDevicePrimaryCtxRelease ( CUdevice dev )
    "cuDevicePrimaryCtxRelease": (c_int, cu_device),
    # CUresult cuDevicePrimaryCtxReset ( CUdevice dev )
    "cuDevicePrimaryCtxReset": (c_int, cu_device),
    # CUresult cuDevicePrimaryCtxRetain ( CUcontext* pctx, CUdevice dev )
    "cuDevicePrimaryCtxRetain": (c_int, POINTER(cu_context), cu_device),
    # CUresult cuDevicePrimaryCtxSetFlags ( CUdevice dev, unsigned int  flags )
    "cuDevicePrimaryCtxSetFlags": (c_int, cu_device, c_uint),
    # CUresult cuCtxCreate(CUcontext *pctx, unsigned int flags,
    #                      CUdevice dev);
    "cuCtxCreate": (c_int, POINTER(cu_context), c_uint, cu_device),
    # CUresult cuCtxGetDevice (	CUdevice * 	device	 )
    "cuCtxGetDevice": (c_int, POINTER(cu_device)),
    # CUresult cuCtxGetCurrent (CUcontext *pctx);
    "cuCtxGetCurrent": (c_int, POINTER(cu_context)),
    # CUresult cuCtxPushCurrent (CUcontext pctx);
    "cuCtxPushCurrent": (c_int, cu_context),
    # CUresult cuCtxPopCurrent (CUcontext *pctx);
    "cuCtxPopCurrent": (c_int, POINTER(cu_context)),
    # CUresult cuCtxDestroy(CUcontext pctx);
    "cuCtxDestroy": (c_int, cu_context),
    # CUresult cuModuleLoadDataEx(CUmodule *module, const void *image,
    #                             unsigned int numOptions,
    #                             CUjit_option *options,
    #                             void **optionValues);
    "cuModuleLoadDataEx": (
        c_int,
        cu_module,
        c_void_p,
        c_uint,
        POINTER(cu_jit_option),
        POINTER(c_void_p),
    ),
    # CUresult cuModuleUnload(CUmodule hmod);
    "cuModuleUnload": (c_int, cu_module),
    # CUresult cuModuleGetFunction(CUfunction *hfunc, CUmodule hmod,
    #                              const char *name);
    "cuModuleGetFunction": (c_int, cu_function, cu_module, c_char_p),
    # CUresult cuModuleGetGlobal ( CUdeviceptr* dptr, size_t* bytes, CUmodule
    #                              hmod, const char* name )
    "cuModuleGetGlobal": (
        c_int,
        POINTER(cu_device_ptr),
        POINTER(c_size_t),
        cu_module,
        c_char_p,
    ),
    # CUresult CUDAAPI cuFuncSetCacheConfig(CUfunction hfunc,
    #                                       CUfunc_cache config);
    "cuFuncSetCacheConfig": (c_int, cu_function, c_uint),
    # CUresult cuMemAlloc(CUdeviceptr *dptr, size_t bytesize);
    "cuMemAlloc": (c_int, POINTER(cu_device_ptr), c_size_t),
    # CUresult cuMemAllocManaged(CUdeviceptr *dptr, size_t bytesize,
    #                            unsigned int flags);
    "cuMemAllocManaged": (c_int, c_void_p, c_size_t, c_uint),
    # CUresult cuMemsetD8(CUdeviceptr dstDevice, unsigned char uc, size_t N)
    "cuMemsetD8": (c_int, cu_device_ptr, c_uint8, c_size_t),
    # CUresult cuMemsetD8Async(CUdeviceptr dstDevice, unsigned char uc,
    #                          size_t N, CUstream hStream);
    "cuMemsetD8Async": (c_int, cu_device_ptr, c_uint8, c_size_t, cu_stream),
    # CUresult cuMemcpyHtoD(CUdeviceptr dstDevice, const void *srcHost,
    #                       size_t ByteCount);
    "cuMemcpyHtoD": (c_int, cu_device_ptr, c_void_p, c_size_t),
    # CUresult cuMemcpyHtoDAsync(CUdeviceptr dstDevice, const void *srcHost,
    #                            size_t ByteCount, CUstream hStream);
    "cuMemcpyHtoDAsync": (c_int, cu_device_ptr, c_void_p, c_size_t, cu_stream),
    # CUresult cuMemcpyDtoD(CUdeviceptr dstDevice, const void *srcDevice,
    #                       size_t ByteCount);
    "cuMemcpyDtoD": (c_int, cu_device_ptr, cu_device_ptr, c_size_t),
    # CUresult cuMemcpyDtoDAsync(CUdeviceptr dstDevice, const void *srcDevice,
    #                            size_t ByteCount, CUstream hStream);
    "cuMemcpyDtoDAsync": (
        c_int,
        cu_device_ptr,
        cu_device_ptr,
        c_size_t,
        cu_stream,
    ),
    # CUresult cuMemcpyDtoH(void *dstHost, CUdeviceptr srcDevice,
    #                       size_t ByteCount);
    "cuMemcpyDtoH": (c_int, c_void_p, cu_device_ptr, c_size_t),
    # CUresult cuMemcpyDtoHAsync(void *dstHost, CUdeviceptr srcDevice,
    #                            size_t ByteCount, CUstream hStream);
    "cuMemcpyDtoHAsync": (c_int, c_void_p, cu_device_ptr, c_size_t, cu_stream),
    # CUresult cuMemFree(CUdeviceptr dptr);
    "cuMemFree": (c_int, cu_device_ptr),
    # CUresult cuStreamCreate(CUstream *phStream, unsigned int Flags);
    "cuStreamCreate": (c_int, POINTER(cu_stream), c_uint),
    # CUresult cuStreamDestroy(CUstream hStream);
    "cuStreamDestroy": (c_int, cu_stream),
    # CUresult cuStreamSynchronize(CUstream hStream);
    "cuStreamSynchronize": (c_int, cu_stream),
    # CUresult cuStreamAddCallback(
    #              CUstream hStream,
    #              CUstreamCallback callback,
    #              void* userData,
    #              unsigned int flags)
    "cuStreamAddCallback": (
        c_int,
        cu_stream,
        cu_stream_callback_pyobj,
        py_object,
        c_uint,
    ),
    # CUresult cuLaunchKernel(CUfunction f, unsigned int gridDimX,
    #                        unsigned int gridDimY,
    #                        unsigned int gridDimZ,
    #                        unsigned int blockDimX,
    #                        unsigned int blockDimY,
    #                        unsigned int blockDimZ,
    #                        unsigned int sharedMemBytes,
    #                        CUstream hStream, void **kernelParams,
    #                        void ** extra)
    "cuLaunchKernel": (
        c_int,
        cu_function,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        cu_stream,
        POINTER(c_void_p),
        POINTER(c_void_p),
    ),
    # CUresult cuLaunchCooperativeKernel(CUfunction f, unsigned int gridDimX,
    #                                   unsigned int gridDimY,
    #                                   unsigned int gridDimZ,
    #                                   unsigned int blockDimX,
    #                                   unsigned int blockDimY,
    #                                   unsigned int blockDimZ,
    #                                   unsigned int sharedMemBytes,
    #                                   CUstream hStream, void **kernelParams)
    "cuLaunchCooperativeKernel": (
        c_int,
        cu_function,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        c_uint,
        cu_stream,
        POINTER(c_void_p),
    ),
    #  CUresult cuMemHostAlloc (	void ** 	pp,
    #                               size_t 	bytesize,
    #                               unsigned int 	Flags
    #                           )
    "cuMemHostAlloc": (c_int, c_void_p, c_size_t, c_uint),
    #  CUresult cuMemFreeHost (	void * 	p	 )
    "cuMemFreeHost": (c_int, c_void_p),
    # CUresult cuMemHostRegister(void * 	p,
    #                            size_t 	bytesize,
    #                            unsigned int 	Flags)
    "cuMemHostRegister": (c_int, c_void_p, c_size_t, c_uint),
    # CUresult cuMemHostUnregister(void * 	p)
    "cuMemHostUnregister": (c_int, c_void_p),
    # CUresult cuMemHostGetDevicePointer(CUdeviceptr * pdptr,
    #                                    void *        p,
    #                                    unsigned int  Flags)
    "cuMemHostGetDevicePointer": (
        c_int,
        POINTER(cu_device_ptr),
        c_void_p,
        c_uint,
    ),
    # CUresult cuMemGetInfo(size_t * free, size_t * total)
    "cuMemGetInfo": (c_int, POINTER(c_size_t), POINTER(c_size_t)),
    # CUresult cuEventCreate (	CUevent * 	phEvent,
    #                               unsigned int 	Flags )
    "cuEventCreate": (c_int, POINTER(cu_event), c_uint),
    # CUresult cuEventDestroy (	CUevent 	hEvent	 )
    "cuEventDestroy": (c_int, cu_event),
    # CUresult cuEventElapsedTime (	float * 	pMilliseconds,
    #                                   CUevent 	hStart,
    #                                   CUevent 	hEnd )
    "cuEventElapsedTime": (c_int, POINTER(c_float), cu_event, cu_event),
    # CUresult cuEventQuery (	CUevent 	hEvent	 )
    "cuEventQuery": (c_int, cu_event),
    # CUresult cuEventRecord (	CUevent 	hEvent,
    #                               CUstream 	hStream )
    "cuEventRecord": (c_int, cu_event, cu_stream),
    # CUresult cuEventSynchronize (	CUevent 	hEvent	 )
    "cuEventSynchronize": (c_int, cu_event),
    # CUresult cuStreamWaitEvent (	CUstream        hStream,
    #                                   CUevent         hEvent,
    #                                	unsigned int 	Flags )
    "cuStreamWaitEvent": (c_int, cu_stream, cu_event, c_uint),
    # CUresult 	cuPointerGetAttribute (
    #               void *data,
    #               CUpointer_attribute attribute,
    #               CUdeviceptr ptr)
    "cuPointerGetAttribute": (c_int, c_void_p, c_uint, cu_device_ptr),
    #    CUresult cuMemGetAddressRange (	CUdeviceptr * 	pbase,
    #                                        size_t * 	psize,
    #                                        CUdeviceptr 	dptr
    #                                        )
    "cuMemGetAddressRange": (
        c_int,
        POINTER(cu_device_ptr),
        POINTER(c_size_t),
        cu_device_ptr,
    ),
    #    CUresult cuMemHostGetFlags (	unsigned int * 	pFlags,
    #                                   void * 	p )
    "cuMemHostGetFlags": (c_int, POINTER(c_uint), c_void_p),
    #   CUresult cuCtxSynchronize ( void )
    "cuCtxSynchronize": (c_int,),
    #    CUresult
    #    cuLinkCreate(unsigned int numOptions, CUjit_option *options,
    #                 void **optionValues, CUlinkState *stateOut);
    "cuLinkCreate": (
        c_int,
        c_uint,
        POINTER(cu_jit_option),
        POINTER(c_void_p),
        POINTER(cu_link_state),
    ),
    #    CUresult
    #    cuLinkAddData(CUlinkState state, CUjitInputType type, void *data,
    #                  size_t size, const char *name, unsigned
    #                  int numOptions, CUjit_option *options,
    #                  void **optionValues);
    "cuLinkAddData": (
        c_int,
        cu_link_state,
        cu_jit_input_type,
        c_void_p,
        c_size_t,
        c_char_p,
        c_uint,
        POINTER(cu_jit_option),
        POINTER(c_void_p),
    ),
    #    CUresult
    #    cuLinkAddFile(CUlinkState state, CUjitInputType type,
    #                  const char *path, unsigned int numOptions,
    #                  CUjit_option *options, void **optionValues);
    "cuLinkAddFile": (
        c_int,
        cu_link_state,
        cu_jit_input_type,
        c_char_p,
        c_uint,
        POINTER(cu_jit_option),
        POINTER(c_void_p),
    ),
    #    CUresult CUDAAPI
    #    cuLinkComplete(CUlinkState state, void **cubinOut, size_t *sizeOut)
    "cuLinkComplete": (
        c_int,
        cu_link_state,
        POINTER(c_void_p),
        POINTER(c_size_t),
    ),
    #    CUresult CUDAAPI
    #    cuLinkDestroy(CUlinkState state)
    "cuLinkDestroy": (c_int, cu_link_state),
    # cuProfilerStart ( void )
    "cuProfilerStart": (c_int,),
    # cuProfilerStop ( void )
    "cuProfilerStop": (c_int,),
    # CUresult cuFuncGetAttribute ( int* pi, CUfunction_attribute attrib,
    #                              CUfunction hfunc )
    "cuFuncGetAttribute": (
        c_int,
        POINTER(c_int),
        cu_function_attribute,
        cu_function,
    ),
    # CUresult CUDAAPI cuOccupancyMaxActiveBlocksPerMultiprocessor(
    #                      int *numBlocks,
    #                      CUfunction func,
    #                      int blockSize,
    #                      size_t dynamicSMemSize);
    "cuOccupancyMaxActiveBlocksPerMultiprocessor": (
        c_int,
        POINTER(c_int),
        cu_function,
        c_size_t,
        c_uint,
    ),
    # CUresult CUDAAPI cuOccupancyMaxActiveBlocksPerMultiprocessorWithFlags(
    #                      int *numBlocks,
    #                      CUfunction func,
    #                      int blockSize,
    #                      size_t dynamicSMemSize,
    #                      unsigned int flags);
    "cuOccupancyMaxActiveBlocksPerMultiprocessorWithFlags": (
        c_int,
        POINTER(c_int),
        cu_function,
        c_size_t,
        c_uint,
    ),
    # CUresult CUDAAPI cuOccupancyMaxPotentialBlockSize(
    #                      int *minGridSize, int *blockSize,
    #                      CUfunction func,
    #                      CUoccupancyB2DSize blockSizeToDynamicSMemSize,
    #                      size_t dynamicSMemSize, int blockSizeLimit);
    "cuOccupancyMaxPotentialBlockSize": (
        c_int,
        POINTER(c_int),
        POINTER(c_int),
        cu_function,
        cu_occupancy_b2d_size,
        c_size_t,
        c_int,
    ),
    # CUresult CUDAAPI cuOccupancyMaxPotentialBlockSizeWithFlags(
    #                      int *minGridSize, int *blockSize,
    #                      CUfunction func,
    #                      CUoccupancyB2DSize blockSizeToDynamicSMemSize,
    #                      size_t dynamicSMemSize, int blockSizeLimit,
    #                      unsigned int flags);
    "cuOccupancyMaxPotentialBlockSizeWithFlags": (
        c_int,
        POINTER(c_int),
        POINTER(c_int),
        cu_function,
        cu_occupancy_b2d_size,
        c_size_t,
        c_int,
        c_uint,
    ),
    # CUresult cuIpcGetMemHandle ( CUipcMemHandle* pHandle, CUdeviceptr dptr )
    "cuIpcGetMemHandle": (c_int, POINTER(cu_ipc_mem_handle), cu_device_ptr),
    # CUresult cuIpcOpenMemHandle(
    #              CUdeviceptr* pdptr,
    #              CUipcMemHandle handle,
    #              unsigned int Flags)
    "cuIpcOpenMemHandle": (
        c_int,
        POINTER(cu_device_ptr),
        cu_ipc_mem_handle,
        c_uint,
    ),
    # CUresult cuIpcCloseMemHandle ( CUdeviceptr dptr )
    "cuIpcCloseMemHandle": (c_int, cu_device_ptr),
    # CUresult cuCtxEnablePeerAccess (CUcontext peerContext, unsigned int Flags)
    "cuCtxEnablePeerAccess": (c_int, cu_context, c_int),
    # CUresult cuDeviceCanAccessPeer ( int* canAccessPeer,
    #                                  CUdevice dev, CUdevice peerDev )
    "cuDeviceCanAccessPeer": (c_int, POINTER(c_int), cu_device, cu_device),
    # CUresult cuDeviceGetUuid ( CUuuid* uuid, CUdevice dev )
    "cuDeviceGetUuid": (c_int, POINTER(cu_uuid), cu_device),
}



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
NVVM is not supported in the simulator, but stubs are provided to allow tests
to import correctly.
"""


class NvvmSupportError(ImportError):
    pass


class NVVM:
    def __init__(self):
        raise NvvmSupportError("NVVM not supported in the simulator")


CompilationUnit = None
compile_ir = None
set_cuda_kernel = None
get_arch_option = None
LibDevice = None
NvvmError = None


def is_available():
    return False


def get_supported_ccs():
    return ()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
This implements the typing template for `dict()`.
"""

from numba.cuda import types
from numba.cuda.core import errors
from .templates import (
    AbstractTemplate,
    Registry,
    signature,
)

registry = Registry()
infer = registry.register
infer_global = registry.register_global
infer_getattr = registry.register_attr


_message_dict_support = """
Unsupported use of `dict()` with positional or keyword argument(s). \
The only supported uses are `dict()` or `dict(iterable)`.
""".strip()


@infer_global(dict)
class DictBuiltin(AbstractTemplate):
    def generic(self, args, kws):
        if kws:
            raise errors.TypingError(_message_dict_support)
        if args:
            (iterable,) = args
            if isinstance(iterable, types.IterableType):
                dtype = iterable.iterator_type.yield_type
                if isinstance(dtype, types.UniTuple):
                    length = dtype.count
                    if length != 2:
                        msg = (
                            "dictionary update sequence element has length "
                            f"{length}; 2 is required"
                        )
                        raise errors.TypingError(msg)
                    k = v = dtype.key[0]
                elif isinstance(dtype, types.Tuple):
                    k, v = dtype.key
                else:
                    raise errors.TypingError(_message_dict_support)

                # dict key must be hashable
                if not isinstance(k, types.Hashable):
                    msg = f"Unhashable type: {k}"
                    raise errors.TypingError(msg)

                return signature(types.DictType(k, v), iterable)
            else:
                msg = (
                    "Non-iterable args used in dict(iterable) "
                    f"constructor. Got 'dict({args[0]})'"
                )
                raise errors.TypingError(msg)
        return signature(types.DictType(types.undefined, types.undefined))



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from .manager import DataModelManager
from .packer import ArgPacker, DataPacker
from .registry import register_default, default_manager, register
from .models import PrimitiveModel, CompositeModel, StructModel  # type: ignore



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda import config


class RTSys:
    def __init__(self, *args, **kwargs):
        pass

    def memsys_enable_stats(self):
        pass

    def get_allocation_stats(self):
        pass


rtsys = RTSys()

config.CUDA_NRT_STATS = False
config.CUDA_ENABLE_NRT = False



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
import sys

from numba.cuda._internal.cuda_bf16 import (
    typing_registry,
    target_registry,
    nv_bfloat16 as bfloat16,
    # Arithmetic intrinsics
    __habs as habs,
    __hadd as hadd,
    __hsub as hsub,
    __hmul as hmul,
    __hadd_rn as hadd_rn,
    __hsub_rn as hsub_rn,
    __hmul_rn as hmul_rn,
    __hdiv as hdiv,
    __hadd_sat as hadd_sat,
    __hsub_sat as hsub_sat,
    __hmul_sat as hmul_sat,
    __hfma as hfma,
    __hfma_sat as hfma_sat,
    __hneg as hneg,
    __hfma_relu as hfma_relu,
    # Comparison intrinsics
    __heq as heq,
    __hne as hne,
    __hge as hge,
    __hgt as hgt,
    __hle as hle,
    __hlt as hlt,
    __hmax as hmax,
    __hmin as hmin,
    __hmax_nan as hmax_nan,
    __hmin_nan as hmin_nan,
    __hisinf as hisinf,
    __hisnan as hisnan,
    # Unordered comparison intrinsics
    __hequ as hequ,
    __hneu as hneu,
    __hgeu as hgeu,
    __hgtu as hgtu,
    __hleu as hleu,
    __hltu as hltu,
    # Precision conversion and data movement
    # - floating-point family
    __bfloat162float as bfloat162float,
    __float2bfloat16 as float2bfloat16,
    __double2bfloat16 as double2bfloat16,
    __float2bfloat16_rn as float2bfloat16_rn,
    __float2bfloat16_rz as float2bfloat16_rz,
    __float2bfloat16_rd as float2bfloat16_rd,
    __float2bfloat16_ru as float2bfloat16_ru,
    # - char family
    __bfloat162char_rz as bfloat162char_rz,
    __bfloat162uchar_rz as bfloat162uchar_rz,
    # - int family (signed 32-bit)
    __int2bfloat16_rn as int2bfloat16_rn,
    __int2bfloat16_rz as int2bfloat16_rz,
    __int2bfloat16_rd as int2bfloat16_rd,
    __int2bfloat16_ru as int2bfloat16_ru,
    __bfloat162int_rn as bfloat162int_rn,
    __bfloat162int_rz as bfloat162int_rz,
    __bfloat162int_rd as bfloat162int_rd,
    __bfloat162int_ru as bfloat162int_ru,
    # - short family (signed 16-bit)
    __short2bfloat16_rn as short2bfloat16_rn,
    __short2bfloat16_rz as short2bfloat16_rz,
    __short2bfloat16_rd as short2bfloat16_rd,
    __short2bfloat16_ru as short2bfloat16_ru,
    __bfloat162short_rn as bfloat162short_rn,
    __bfloat162short_rz as bfloat162short_rz,
    __bfloat162short_rd as bfloat162short_rd,
    __bfloat162short_ru as bfloat162short_ru,
    # - ushort family (unsigned 16-bit)
    __ushort2bfloat16_rn as ushort2bfloat16_rn,
    __ushort2bfloat16_rz as ushort2bfloat16_rz,
    __ushort2bfloat16_rd as ushort2bfloat16_rd,
    __ushort2bfloat16_ru as ushort2bfloat16_ru,
    __bfloat162ushort_rn as bfloat162ushort_rn,
    __bfloat162ushort_rz as bfloat162ushort_rz,
    __bfloat162ushort_rd as bfloat162ushort_rd,
    __bfloat162ushort_ru as bfloat162ushort_ru,
    # - uint family (unsigned 32-bit)
    __uint2bfloat16_rn as uint2bfloat16_rn,
    __uint2bfloat16_rz as uint2bfloat16_rz,
    __uint2bfloat16_rd as uint2bfloat16_rd,
    __uint2bfloat16_ru as uint2bfloat16_ru,
    __bfloat162uint_rn as bfloat162uint_rn,
    __bfloat162uint_rz as bfloat162uint_rz,
    __bfloat162uint_rd as bfloat162uint_rd,
    __bfloat162uint_ru as bfloat162uint_ru,
    # - ll family (signed 64-bit)
    __ll2bfloat16_rn as ll2bfloat16_rn,
    __ll2bfloat16_rz as ll2bfloat16_rz,
    __ll2bfloat16_rd as ll2bfloat16_rd,
    __ll2bfloat16_ru as ll2bfloat16_ru,
    __bfloat162ll_rn as bfloat162ll_rn,
    __bfloat162ll_rz as bfloat162ll_rz,
    __bfloat162ll_rd as bfloat162ll_rd,
    __bfloat162ll_ru as bfloat162ll_ru,
    # - ull family (unsigned 64-bit)
    __ull2bfloat16_rn as ull2bfloat16_rn,
    __ull2bfloat16_rz as ull2bfloat16_rz,
    __ull2bfloat16_rd as ull2bfloat16_rd,
    __ull2bfloat16_ru as ull2bfloat16_ru,
    __bfloat162ull_rn as bfloat162ull_rn,
    __bfloat162ull_rz as bfloat162ull_rz,
    __bfloat162ull_rd as bfloat162ull_rd,
    __bfloat162ull_ru as bfloat162ull_ru,
    # - bit reinterpret casts
    __bfloat16_as_short as bfloat16_as_short,
    __bfloat16_as_ushort as bfloat16_as_ushort,
    __short_as_bfloat16 as short_as_bfloat16,
    __ushort_as_bfloat16 as ushort_as_bfloat16,
    htrunc,
    hceil,
    hfloor,
    hrint,
    hsqrt,
    hrsqrt,
    hrcp,
    hlog,
    hlog2,
    hlog10,
    hcos,
    hsin,
    hexp,
    hexp2,
    hexp10,
    htanh,
    htanh_approx,
)
from numba.cuda.extending import overload

import math


def _make_unary(a, func):
    if a == bfloat16:
        return lambda a: func(a)


# Bind low++ bindings to math APIs
@overload(math.trunc, target="cuda")
def trunc_ol(a):
    return _make_unary(a, htrunc)


@overload(math.ceil, target="cuda")
def ceil_ol(a):
    return _make_unary(a, hceil)


@overload(math.floor, target="cuda")
def floor_ol(a):
    return _make_unary(a, hfloor)


@overload(math.sqrt, target="cuda")
def sqrt_ol(a):
    return _make_unary(a, hsqrt)


@overload(math.log, target="cuda")
def log_ol(a):
    return _make_unary(a, hlog)


@overload(math.log10, target="cuda")
def log10_ol(a):
    return _make_unary(a, hlog10)


@overload(math.cos, target="cuda")
def cos_ol(a):
    return _make_unary(a, hcos)


@overload(math.sin, target="cuda")
def sin_ol(a):
    return _make_unary(a, hsin)


@overload(math.tanh, target="cuda")
def tanh_ol(a):
    return _make_unary(a, htanh)


@overload(math.exp, target="cuda")
def exp_ol(a):
    return _make_unary(a, hexp)


if sys.version_info >= (3, 11):

    @overload(math.exp2, target="cuda")
    def exp2_ol(a):
        return _make_unary(a, hexp2)


## Public aliases using Numba/Numpy-style type names
# Floating-point
float32_to_bfloat16 = float2bfloat16
float64_to_bfloat16 = double2bfloat16
bfloat16_to_float32 = bfloat162float
float32_to_bfloat16_rn = float2bfloat16_rn
float32_to_bfloat16_rz = float2bfloat16_rz
float32_to_bfloat16_rd = float2bfloat16_rd
float32_to_bfloat16_ru = float2bfloat16_ru

# Char (8-bit)
bfloat16_to_int8_rz = bfloat162char_rz
bfloat16_to_uint8_rz = bfloat162uchar_rz

# Int16 / UInt16
int16_to_bfloat16_rn = short2bfloat16_rn
int16_to_bfloat16_rz = short2bfloat16_rz
int16_to_bfloat16_rd = short2bfloat16_rd
int16_to_bfloat16_ru = short2bfloat16_ru
bfloat16_to_int16_rn = bfloat162short_rn
bfloat16_to_int16_rz = bfloat162short_rz
bfloat16_to_int16_rd = bfloat162short_rd
bfloat16_to_int16_ru = bfloat162short_ru

uint16_to_bfloat16_rn = ushort2bfloat16_rn
uint16_to_bfloat16_rz = ushort2bfloat16_rz
uint16_to_bfloat16_rd = ushort2bfloat16_rd
uint16_to_bfloat16_ru = ushort2bfloat16_ru
bfloat16_to_uint16_rn = bfloat162ushort_rn
bfloat16_to_uint16_rz = bfloat162ushort_rz
bfloat16_to_uint16_rd = bfloat162ushort_rd
bfloat16_to_uint16_ru = bfloat162ushort_ru

# Int32 / UInt32
int32_to_bfloat16_rn = int2bfloat16_rn
int32_to_bfloat16_rz = int2bfloat16_rz
int32_to_bfloat16_rd = int2bfloat16_rd
int32_to_bfloat16_ru = int2bfloat16_ru
bfloat16_to_int32_rn = bfloat162int_rn
bfloat16_to_int32_rz = bfloat162int_rz
bfloat16_to_int32_rd = bfloat162int_rd
bfloat16_to_int32_ru = bfloat162int_ru

uint32_to_bfloat16_rn = uint2bfloat16_rn
uint32_to_bfloat16_rz = uint2bfloat16_rz
uint32_to_bfloat16_rd = uint2bfloat16_rd
uint32_to_bfloat16_ru = uint2bfloat16_ru
bfloat16_to_uint32_rn = bfloat162uint_rn
bfloat16_to_uint32_rz = bfloat162uint_rz
bfloat16_to_uint32_rd = bfloat162uint_rd
bfloat16_to_uint32_ru = bfloat162uint_ru

# Int64 / UInt64
int64_to_bfloat16_rn = ll2bfloat16_rn
int64_to_bfloat16_rz = ll2bfloat16_rz
int64_to_bfloat16_rd = ll2bfloat16_rd
int64_to_bfloat16_ru = ll2bfloat16_ru
bfloat16_to_int64_rn = bfloat162ll_rn
bfloat16_to_int64_rz = bfloat162ll_rz
bfloat16_to_int64_rd = bfloat162ll_rd
bfloat16_to_int64_ru = bfloat162ll_ru

uint64_to_bfloat16_rn = ull2bfloat16_rn
uint64_to_bfloat16_rz = ull2bfloat16_rz
uint64_to_bfloat16_rd = ull2bfloat16_rd
uint64_to_bfloat16_ru = ull2bfloat16_ru
bfloat16_to_uint64_rn = bfloat162ull_rn
bfloat16_to_uint64_rz = bfloat162ull_rz
bfloat16_to_uint64_rd = bfloat162ull_rd
bfloat16_to_uint64_ru = bfloat162ull_ru

# Bit reinterpret casts
bfloat16_as_int16 = bfloat16_as_short
bfloat16_as_uint16 = bfloat16_as_ushort
int16_as_bfloat16 = short_as_bfloat16
uint16_as_bfloat16 = ushort_as_bfloat16

__all__ = [
    "typing_registry",
    "target_registry",
    "bfloat16",
    # Arithmetic intrinsics
    "habs",
    "hadd",
    "hsub",
    "hmul",
    "hadd_rn",
    "hsub_rn",
    "hmul_rn",
    "hdiv",
    "hadd_sat",
    "hsub_sat",
    "hmul_sat",
    "hfma",
    "hfma_sat",
    "hneg",
    "hfma_relu",
    # Comparison intrinsics
    "heq",
    "hne",
    "hge",
    "hgt",
    "hle",
    "hlt",
    "hmax",
    "hmin",
    "hmax_nan",
    "hmin_nan",
    "hisinf",
    "hisnan",
    "hequ",
    "hneu",
    "hgeu",
    "hgtu",
    "hleu",
    "hltu",
    # Precision conversion and data movement
    "float32_to_bfloat16",
    "float64_to_bfloat16",
    "bfloat16_to_float32",
    "float32_to_bfloat16_rn",
    "float32_to_bfloat16_rz",
    "float32_to_bfloat16_rd",
    "float32_to_bfloat16_ru",
    "bfloat16_to_int8_rz",
    "bfloat16_to_uint8_rz",
    "int16_to_bfloat16_rn",
    "int16_to_bfloat16_rz",
    "int16_to_bfloat16_rd",
    "int16_to_bfloat16_ru",
    "bfloat16_to_int16_rn",
    "bfloat16_to_int16_rz",
    "bfloat16_to_int16_rd",
    "bfloat16_to_int16_ru",
    "uint16_to_bfloat16_rn",
    "uint16_to_bfloat16_rz",
    "uint16_to_bfloat16_rd",
    "uint16_to_bfloat16_ru",
    "bfloat16_to_uint16_rn",
    "bfloat16_to_uint16_rz",
    "bfloat16_to_uint16_rd",
    "bfloat16_to_uint16_ru",
    "int32_to_bfloat16_rn",
    "int32_to_bfloat16_rz",
    "int32_to_bfloat16_rd",
    "int32_to_bfloat16_ru",
    "bfloat16_to_int32_rn",
    "bfloat16_to_int32_rz",
    "bfloat16_to_int32_rd",
    "bfloat16_to_int32_ru",
    "uint32_to_bfloat16_rn",
    "uint32_to_bfloat16_rz",
    "uint32_to_bfloat16_rd",
    "uint32_to_bfloat16_ru",
    "bfloat16_to_uint32_rn",
    "bfloat16_to_uint32_rz",
    "bfloat16_to_uint32_rd",
    "bfloat16_to_uint32_ru",
    "int64_to_bfloat16_rn",
    "int64_to_bfloat16_rz",
    "int64_to_bfloat16_rd",
    "int64_to_bfloat16_ru",
    "bfloat16_to_int64_rn",
    "bfloat16_to_int64_rz",
    "bfloat16_to_int64_rd",
    "bfloat16_to_int64_ru",
    "uint64_to_bfloat16_rn",
    "uint64_to_bfloat16_rz",
    "uint64_to_bfloat16_rd",
    "uint64_to_bfloat16_ru",
    "bfloat16_to_uint64_rn",
    "bfloat16_to_uint64_rz",
    "bfloat16_to_uint64_rd",
    "bfloat16_to_uint64_ru",
    "bfloat16_as_int16",
    "bfloat16_as_uint16",
    "int16_as_bfloat16",
    "uint16_as_bfloat16",
    "htrunc",
    "hceil",
    "hfloor",
    "hrint",
    "hsqrt",
    "hrsqrt",
    "hrcp",
    "hlog",
    "hlog2",
    "hlog10",
    "hcos",
    "hsin",
    "htanh",
    "htanh_approx",
    "hexp",
    "hexp2",
    "hexp10",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from functools import reduce as pyreduce


def Reduce(func):
    def reduce_wrapper(seq, res=None, init=0):
        r = pyreduce(func, seq, init)
        if res is not None:
            res[0] = r
            return None
        else:
            return r

    return reduce_wrapper


reduce = Reduce



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
NumPy extensions.
"""

from numba.cuda.np.arraymath import cross2d


__all__ = ["cross2d"]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Support for CFFI. Allows checking whether objects are CFFI functions and
obtaining the pointer and numba signature.
"""

from functools import partial
import numpy as np

from numba.cuda import types
from numba.cuda.core.errors import TypingError
from numba.cuda.typing import templates
from numba.cuda.np import numpy_support

try:
    import cffi

    ffi = cffi.FFI()
except ImportError:
    ffi = None

SUPPORTED = ffi is not None
_ool_func_types = {}
_ool_func_ptr = {}
_ffi_instances = set()


def is_ffi_instance(obj):
    # Compiled FFI modules have a member, ffi, which is an instance of
    # CompiledFFI, which behaves similarly to an instance of cffi.FFI. In
    # order to simplify handling a CompiledFFI object, we treat them as
    # if they're cffi.FFI instances for typing and lowering purposes.
    try:
        return obj in _ffi_instances or isinstance(obj, cffi.FFI)
    except TypeError:  # Unhashable type possible
        return False


def is_cffi_func(obj):
    """Check whether the obj is a CFFI function"""
    try:
        return ffi.typeof(obj).kind == "function"
    except TypeError:
        try:
            return obj in _ool_func_types
        except Exception:
            return False


def get_pointer(cffi_func):
    """
    Get a pointer to the underlying function for a CFFI function as an
    integer.
    """
    if cffi_func in _ool_func_ptr:
        return _ool_func_ptr[cffi_func]
    return int(ffi.cast("uintptr_t", cffi_func))


_cached_type_map = None


def _type_map():
    """
    Lazily compute type map, as calling ffi.typeof() involves costly
    parsing of C code...
    """
    global _cached_type_map
    if _cached_type_map is None:
        _cached_type_map = {
            ffi.typeof("bool"): types.boolean,
            ffi.typeof("char"): types.char,
            ffi.typeof("short"): types.short,
            ffi.typeof("int"): types.intc,
            ffi.typeof("long"): types.long_,
            ffi.typeof("long long"): types.longlong,
            ffi.typeof("unsigned char"): types.uchar,
            ffi.typeof("unsigned short"): types.ushort,
            ffi.typeof("unsigned int"): types.uintc,
            ffi.typeof("unsigned long"): types.ulong,
            ffi.typeof("unsigned long long"): types.ulonglong,
            ffi.typeof("int8_t"): types.char,
            ffi.typeof("uint8_t"): types.uchar,
            ffi.typeof("int16_t"): types.short,
            ffi.typeof("uint16_t"): types.ushort,
            ffi.typeof("int32_t"): types.intc,
            ffi.typeof("uint32_t"): types.uintc,
            ffi.typeof("int64_t"): types.longlong,
            ffi.typeof("uint64_t"): types.ulonglong,
            ffi.typeof("float"): types.float32,
            ffi.typeof("double"): types.double,
            ffi.typeof("ssize_t"): types.intp,
            ffi.typeof("size_t"): types.uintp,
            ffi.typeof("void"): types.void,
        }
    return _cached_type_map


def map_type(cffi_type, use_record_dtype=False):
    """
    Map CFFI type to numba type.

    Parameters
    ----------
    cffi_type:
        The CFFI type to be converted.
    use_record_dtype: bool (default: False)
        When True, struct types are mapped to a NumPy Record dtype.

    """
    primed_map_type = partial(map_type, use_record_dtype=use_record_dtype)
    kind = getattr(cffi_type, "kind", "")
    if kind == "union":
        raise TypeError("No support for CFFI union")
    elif kind == "function":
        if cffi_type.ellipsis:
            raise TypeError("vararg function is not supported")
        restype = primed_map_type(cffi_type.result)
        argtypes = [primed_map_type(arg) for arg in cffi_type.args]
        return templates.signature(restype, *argtypes)
    elif kind == "pointer":
        pointee = cffi_type.item
        if pointee.kind == "void":
            return types.voidptr
        else:
            return types.CPointer(primed_map_type(pointee))
    elif kind == "array":
        dtype = primed_map_type(cffi_type.item)
        nelem = cffi_type.length
        return types.NestedArray(dtype=dtype, shape=(nelem,))
    elif kind == "struct" and use_record_dtype:
        return map_struct_to_record_dtype(cffi_type)
    else:
        result = _type_map().get(cffi_type)
        if result is None:
            raise TypeError(cffi_type)
        return result


def map_struct_to_record_dtype(cffi_type):
    """Convert a cffi type into a NumPy Record dtype"""
    fields = {
        "names": [],
        "formats": [],
        "offsets": [],
        "itemsize": ffi.sizeof(cffi_type),
    }
    is_aligned = True
    for k, v in cffi_type.fields:
        # guard unsupported values
        if v.bitshift != -1:
            msg = "field {!r} has bitshift, this is not supported"
            raise ValueError(msg.format(k))
        if v.flags != 0:
            msg = "field {!r} has flags, this is not supported"
            raise ValueError(msg.format(k))
        if v.bitsize != -1:
            msg = "field {!r} has bitsize, this is not supported"
            raise ValueError(msg.format(k))
        dtype = numpy_support.as_dtype(
            map_type(v.type, use_record_dtype=True),
        )
        fields["names"].append(k)
        fields["formats"].append(dtype)
        fields["offsets"].append(v.offset)
        # Check alignment
        is_aligned &= v.offset % dtype.alignment == 0

    return numpy_support.from_dtype(np.dtype(fields, align=is_aligned))


def make_function_type(cffi_func, use_record_dtype=False):
    """
    Return a Numba type for the given CFFI function pointer.
    """
    cffi_type = _ool_func_types.get(cffi_func) or ffi.typeof(cffi_func)
    if getattr(cffi_type, "kind", "") == "struct":
        raise TypeError("No support for CFFI struct values")
    sig = map_type(cffi_type, use_record_dtype=use_record_dtype)
    return types.ExternalFunctionPointer(sig, get_pointer=get_pointer)


registry = templates.Registry()


@registry.register
class FFI_from_buffer(templates.AbstractTemplate):
    key = "ffi.from_buffer"

    def generic(self, args, kws):
        if kws or len(args) != 1:
            return
        [ary] = args
        if not isinstance(ary, types.Buffer):
            raise TypingError(
                "from_buffer() expected a buffer object, got %s" % (ary,)
            )
        if ary.layout not in ("C", "F"):
            raise TypingError(
                "from_buffer() unsupported on non-contiguous buffers (got %s)"
                % (ary,)
            )
        if ary.layout != "C" and ary.ndim > 1:
            raise TypingError(
                "from_buffer() only supports multidimensional arrays with C layout (got %s)"
                % (ary,)
            )
        ptr = types.CPointer(ary.dtype)
        return templates.signature(ptr, ary)


@registry.register_attr
class FFIAttribute(templates.AttributeTemplate):
    key = types.ffi

    def resolve_from_buffer(self, ffi):
        return types.BoundFunction(FFI_from_buffer, types.ffi)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys

from . import api, args
from .api import *
from .vector_types import vector_types
from .reduction import Reduce
from .cudadrv.devicearray import (
    device_array,
    device_array_like,
    pinned,
    pinned_array,
    pinned_array_like,
    mapped_array,
    to_device,
    auto_device,
)
from .cudadrv import devicearray
from .cudadrv.devices import require_context, gpus
from .cudadrv.devices import get_context as current_context
from .cudadrv.runtime import runtime
from .cudadrv.linkable_code import LinkableCode
from numba.cuda.core import config

reduce = Reduce

# Register simulated vector types as module level variables
for name, svty in vector_types.items():
    setattr(sys.modules[__name__], name, svty)
    for alias in svty.aliases:
        setattr(sys.modules[__name__], alias, svty)
del vector_types, name, svty, alias

# Ensure that any user code attempting to import cudadrv etc. gets the
# simulator's version and not the real version if the simulator is enabled.
if config.ENABLE_CUDASIM:
    import sys
    from numba.cuda.simulator import cudadrv
    from . import dispatcher

    sys.modules["numba.cuda.api"] = api
    sys.modules["numba.cuda.args"] = args
    sys.modules["numba.cuda.cudadrv"] = cudadrv
    sys.modules["numba.cuda.cudadrv.devicearray"] = cudadrv.devicearray
    sys.modules["numba.cuda.cudadrv.devices"] = cudadrv.devices
    sys.modules["numba.cuda.cudadrv.driver"] = cudadrv.driver
    sys.modules["numba.cuda.cudadrv.linkable_code"] = cudadrv.linkable_code
    sys.modules["numba.cuda.cudadrv.runtime"] = cudadrv.runtime
    sys.modules["numba.cuda.cudadrv.drvapi"] = cudadrv.drvapi
    sys.modules["numba.cuda.cudadrv.error"] = cudadrv.error
    sys.modules["numba.cuda.cudadrv.nvvm"] = cudadrv.nvvm
    sys.modules["numba.cuda.dispatcher"] = dispatcher

    from . import bf16, compiler, _internal

    sys.modules["numba.cuda.bf16"] = bf16
    sys.modules["numba.cuda.compiler"] = compiler
    sys.modules["numba.cuda._internal"] = _internal
    sys.modules["numba.cuda._internal.cuda_bf16"] = _internal.cuda_bf16

    from numba.cuda.simulator import memory_management

    sys.modules["numba.cuda.memory_management"] = memory_management
    sys.modules["numba.cuda.memory_management.nrt"] = memory_management.nrt
